<?php

namespace Tests\Feature;

use App\Models\Danger;
use App\Models\DangerLevel;
use App\Models\Evaluation;
use App\Models\Organization;
use App\Models\User;
use Carbon\Carbon;
use Illuminate\Foundation\Testing\RefreshDatabase;
use Tests\TestCase;

class EvaluationsControllerTest extends TestCase
{
    use RefreshDatabase;

    /**
     * Test get all.
     *
     * @return void
     */
    public function testAdminUserCanGetAllEvaluations()
    {
        $user = User::where('role', User::ROLE_ADMIN)->first();
        $response = $this->actingAs($user)->getJson(route('api.evaluations.all'));

        if ($response->exception) {
            dump($response->exception);
        }

        $response->assertOk();

        $evals = Evaluation::with(['organization', 'dangerLevels', 'evaluationMeasures'])->get();

        $this->assertEquals($evals->toArray(), $response->json());
    }

    /**
     * Test get all.
     *
     * @return void
     */
    public function testUserCanOnlyGetEvaluationsFromOwnOrganization()
    {
        $user = User::where('role', User::ROLE_USER)->first();
        $response = $this->actingAs($user)->getJson(route('api.evaluations.all'));

        $response->assertOk();

        $evals = Evaluation::with(['organization', 'dangerLevels', 'evaluationMeasures'])
            ->where('organization_id', $user->organization_id)
            ->get()
        ;

        $this->assertEquals($evals->toArray(), $response->json());
    }

    /**
     * Test get all.
     *
     * @return void
     */
    public function testManagerCanOnlyGetEvaluationsFromOwnOrganization()
    {
        $user = User::where('role', User::ROLE_MANAGER)->first();
        $response = $this->actingAs($user)->getJson(route('api.evaluations.all'));

        $response->assertOk();

        $evals = Evaluation::with(['organization', 'dangerLevels', 'evaluationMeasures'])
            ->where('organization_id', $user->organization_id)
            ->get()
        ;

        $this->assertEquals($evals->toArray(), $response->json());
    }

    /**
     * test delete.
     */
    public function testAdminCanDeleteAnyEvaluation()
    {
        $user = User::where('role', User::ROLE_ADMIN)->first();

        $eval = Evaluation::first();

        $response = $this->actingAs($user)->deleteJson(route('api.evaluations.delete', ['id' => $eval->id]));

        $response->assertStatus(204);
    }

    /**
     * test delete.
     */
    public function testManagerCanDeleteEvaluationInOwnOrganization()
    {
        $user = User::where('role', User::ROLE_MANAGER)->first();

        $eval = Evaluation::where('organization_id', $user->organization_id)->first();

        $response = $this->actingAs($user)->deleteJson(route('api.evaluations.delete', ['id' => $eval->id]));

        $response->assertStatus(204);
    }

    /**
     * test delete.
     */
    public function testManagerCannotDeleteEvaluationInOtherOrganization()
    {
        $user = User::where('role', User::ROLE_MANAGER)->first();

        $eval = Evaluation::where('organization_id', '!=', $user->organization_id)->first();

        $response = $this->actingAs($user)->deleteJson(route('api.evaluations.delete', ['id' => $eval->id]));

        $response->assertStatus(403);
    }

    /**
     * test delete.
     */
    public function testUserCannotDeleteEvaluation()
    {
        $user = User::where('role', User::ROLE_USER)->first();

        $eval = Evaluation::where('organization_id', $user->organization_id)->first();

        $response = $this->actingAs($user)->deleteJson(route('api.evaluations.delete', ['id' => $eval->id]));

        $response->assertStatus(403);
    }

    public function testPostDraftEvaluationShouldCreateDraftEvaluation()
    {
        $user = User::where('role', User::ROLE_MANAGER)->first();

        $nextId = Evaluation::orderBy('id', 'desc')->first()->id + 1;

        $eval = [
            'organization_id' => $user->organization_id,
            'current_step' => 1,
            'status' => 1,
            'draft' => 1,
            'current_step_status' => 1,
            'updated_by' => $user->firstname . ' ' . $user->lastname,
            'author' => $user->firstname . ' ' . $user->lastname,
            'danger_levels' => [
                ['id' => DangerLevel::first()->id],
            ],
        ];

        $response = $this->actingAs($user)->postJson(route('api.evaluations.post'), $eval);
        $response->assertOk();

        $data = $response->json();

        $dangersLevelsEval = DangerLevel::whereHas('evaluations', function ($q) use ($nextId) {
            $q->where('evaluations.id', $nextId);
        })->get();

        $this->assertEquals([
            'id' => $nextId,
            'status' => Evaluation::STATUS_ONGOING,
            'current_step' => 1,
            'current_step_status' => Evaluation::STATUS_ONGOING,
            'author' => $user->firstname . ' ' . $user->lastname,
            'updated_by' => $user->firstname . ' ' . $user->lastname,
            'organization_id' => $user->organization_id,
            'organization' => $user->organization->toArray(),
            'created_at' => Carbon::now()->format('Y-m-d H:i:s'),
            'updated_at' => Carbon::now()->format('Y-m-d H:i:s'),
            'danger_levels' => [$dangersLevelsEval->toArray()],
            'evaluation_measures' => [],
        ], $data);

        $dbEval = Evaluation::with(['organization', 'dangerLevels', 'evaluationMeasures'])->find($nextId);

        $this->assertNotNull($dbEval);

        $this->assertEquals($dbEval->toArray(), $data);
    }

    public function testPostNonDraftEvaluationWithDataShouldMoveToStep2()
    {
        $user = User::where('role', User::ROLE_MANAGER)->first();

        $nextId = Evaluation::orderBy('id', 'desc')->first()->id + 1;

        $dangerLevels = Danger::all()->map(function (Danger $danger) {
            return $danger->levels()->inRandomOrder()->first()->toArray();
        })->toArray();

        $eval = [
            'organization_id' => $user->organization_id,
            'current_step' => 1,
            'status' => 1,
            'draft' => 0,
            'current_step_status' => 1,
            'updated_by' => $user->firstname . ' ' . $user->lastname,
            'author' => $user->firstname . ' ' . $user->lastname,
            'danger_levels' => $dangerLevels,
        ];

        $response = $this->actingAs($user)->postJson(route('api.evaluations.post'), $eval);
        $response->assertOk();

        $data = $response->json();

        $dangersLevelsEval = DangerLevel::whereHas('evaluations', function ($q) use ($nextId) {
            $q->where('evaluations.id', $nextId);
        })->get();

        $this->assertEquals([
            'id' => $nextId,
            'status' => Evaluation::STATUS_ONGOING,
            'current_step' => 2,
            'current_step_status' => Evaluation::STATUS_ONGOING,
            'author' => $user->firstname . ' ' . $user->lastname,
            'updated_by' => $user->firstname . ' ' . $user->lastname,
            'organization_id' => $user->organization_id,
            'organization' => $user->organization->toArray(),
            'created_at' => Carbon::now()->format('Y-m-d H:i:s'),
            'updated_at' => Carbon::now()->format('Y-m-d H:i:s'),
            'danger_levels' => [$dangersLevelsEval->toArray()],
            'evaluation_measures' => [],
        ], $data);

        $dbEval = Evaluation::with(['organization', 'dangerLevels', 'evaluationMeasures'])->find($nextId);

        $this->assertNotNull($dbEval);

        $this->assertEquals($dbEval->toArray(), $data);
    }

    public function testPostNonDraftEvaluationWithMissingDataDataShouldThrowError()
    {
        $user = User::where('role', User::ROLE_MANAGER)->first();

        $eval = [
            'organization_id' => $user->organization_id,
            'current_step' => 1,
            'status' => 1,
            'draft' => 0,
            'current_step_status' => 1,
            'updated_by' => $user->firstname . ' ' . $user->lastname,
            'author' => $user->firstname . ' ' . $user->lastname,
            'danger_levels' => [
                ['id' => DangerLevel::first()->id],
            ],
        ];

        $response = $this->actingAs($user)->postJson(route('api.evaluations.post'), $eval);
        $response->assertStatus(422);
    }

    public function testPutDraftEvaluationShouldSaveDraftEvaluation()
    {
        $user = User::where('role', User::ROLE_MANAGER)->first();

        $eval = Evaluation::where('organization_id', $user->organization_id)->first()->toArray();

        ++$eval['current_step'];
        $eval['draft'] = 1;

        $response = $this->actingAs($user)->putJson(route('api.evaluations.put', ['id' => $eval['id']]), $eval);
        $response->assertOk();

        $data = $response->json();

        $this->assertEquals([
            'id' => $eval['id'],
            'status' => Evaluation::STATUS_ONGOING,
            'current_step' => $eval['current_step'],
            'current_step_status' => Evaluation::STATUS_ONGOING,
            'author' => $user->firstname . ' ' . $user->lastname,
            'updated_by' => $user->firstname . ' ' . $user->lastname,
            'organization_id' => $user->organization_id,
            'organization' => $user->organization->toArray(),
            'created_at' => Carbon::now()->format('Y-m-d H:i:s'),
            'updated_at' => Carbon::now()->format('Y-m-d H:i:s'),
            'danger_levels' => [],
            'evaluation_measures' => [],
        ], $data);

        $dbEval = Evaluation::with(['organization', 'dangerLevels', 'evaluationMeasures'])->find($eval['id']);

        $this->assertNotNull($dbEval);

        $this->assertEquals($dbEval->toArray(), $data);
    }

    public function testPutEvaluationShouldSaveEvaluationWithFinishedStep()
    {
        $user = User::where('role', User::ROLE_MANAGER)->first();

        $eval = Evaluation::where('organization_id', $user->organization_id)
            ->where('current_step', '<', 5)
            ->first()
            ->toArray();

        $eval['draft'] = 0;

        $response = $this->actingAs($user)->putJson(route('api.evaluations.put', ['id' => $eval['id']]), $eval);
        $response->assertOk();

        $data = $response->json();

        $this->assertEquals([
            'id' => $eval['id'],
            'status' => Evaluation::STATUS_DONE,
            'current_step' => $eval['current_step'],
            'current_step_status' => Evaluation::STATUS_DONE,
            'author' => $user->firstname . ' ' . $user->lastname,
            'updated_by' => $user->firstname . ' ' . $user->lastname,
            'organization_id' => $user->organization_id,
            'organization' => $user->organization->toArray(),
            'created_at' => Carbon::now()->format('Y-m-d H:i:s'),
            'updated_at' => Carbon::now()->format('Y-m-d H:i:s'),
            'danger_levels' => [],
            'evaluation_measures' => [],
        ], $data);

        $dbEval = Evaluation::with(['organization', 'dangerLevels', 'evaluationMeasures'])->find($eval['id']);

        $this->assertNotNull($dbEval);

        $this->assertEquals($dbEval->toArray(), $data);
    }

    public function testManagerCannotCreateEvaluationInOtherOrganization()
    {
        $user = User::where('role', User::ROLE_MANAGER)->first();

        $org = Organization::where('id', '!=', $user->organization_id)->inRandomORder()->first();

        $eval = [
            'organization_id' => $org->id,
            'current_step' => 1,
            'status' => 1,
            'current_step_status' => 1,
            'updated_by' => $user->firstname . ' ' . $user->lastname,
            'author' => $user->firstname . ' ' . $user->lastname,
        ];

        $response = $this->actingAs($user)->postJson(route('api.evaluations.post'), $eval);
        $response->assertStatus(403);
    }

    public function testUserCannotCreateEvaluation()
    {
        $user = User::where('role', User::ROLE_USER)->first();

        $eval = [
            'organization_id' => $user->organization_id,
            'current_step' => 1,
            'status' => 1,
            'current_step_status' => 1,
            'updated_by' => $user->firstname . ' ' . $user->lastname,
            'author' => $user->firstname . ' ' . $user->lastname,
        ];

        $response = $this->actingAs($user)->postJson(route('api.evaluations.post'), $eval);
        $response->assertStatus(403);
    }
}
